#ifndef __LINE_H_
 #define __LINE_H_

#define DEFAULT_BUFFER_SIZE 1000
#define DEFAULT_BUFFER_ADD  500

typedef struct __line_ {
  char *data;
  unsigned long bytes, size, grow;
} line_t;

line_t *line_alloc(void);
int line_init(line_t *, char *, unsigned long);
int line_inject(line_t *, char *, unsigned long);
void line_restart(line_t *);
void line_kill(line_t *);

#ifdef LINE_DEBUG
void line_print(char *, unsigned long);
#endif

#endif
